#ifndef __JABYENGINE_FILE_TYPES_HPP__
#define __JABYENGINE_FILE_TYPES_HPP__
#include "../Auxiliary/bits.hpp"
#include "../GPU/gpu_types.hpp"
#include "../jabyengine_defines.h"

namespace JabyEngine {
    #pragma pack(push, 1)
    struct Nothing {
    };

    struct SimpleTIM {
        static constexpr auto TextureX = BitRange::from_to(0, 8);
        static constexpr auto TextureY = BitRange::from_to(9, 16);
        static constexpr auto ClutX    = BitRange::from_to(17, 22);
        static constexpr auto ClutY    = BitRange::from_to(23, 31);

        uint32_t raw;

        constexpr SimpleTIM() : raw(0) {}

        constexpr SimpleTIM(uint16_t texX, uint16_t texY, uint16_t clutX, uint16_t clutY) : raw(TextureX.as_value(texX >> 1) | TextureY.as_value(texY >> 1) | ClutX.as_value(clutX >> 4) | ClutY.as_value(static_cast<uint32_t>(clutY))) {
        }

        constexpr uint16_t get_texture_x() const {
            return bit::value::get_normalized(this->raw, TextureX) << 1;
        }

        constexpr uint16_t get_texture_y() const {
            return bit::value::get_normalized(this->raw, TextureY) << 1;
        }

        constexpr GPU::PositionU16 get_texture_position() const {
            return {SimpleTIM::get_texture_x(), SimpleTIM::get_texture_y()};
        }

        constexpr uint16_t get_clut_x() const {
            return bit::value::get_normalized(this->raw, SimpleTIM::ClutX) << 4;
        }

        constexpr uint16_t get_clut_y() const {
            return bit::value::get_normalized(this->raw, ClutY);
        }

        constexpr GPU::PositionU16 get_clut_position() const {
            return {SimpleTIM::get_clut_x(), SimpleTIM::get_clut_y()};
        }
    };

    struct CopyTo {
        uint32_t* dst;
    };
    #pragma pack(pop)

    typedef CopyTo Overlay;
}
#endif // !__JABYENGINE_FILE_TYPES_HPP__